.FP H


.ce 100
.B "REAL Users Manual"
.ce0
.sp 2
.fo ''REAL Version 4.0'\*(td'
.he ''%''
.PP
This manual 
provides a tutorial introduction to setting up a simulation in REAL.
If you want to modify REAL you should read also 
read the REAL Programmers Manual.
.NH 1 
General Orientation
.PP
REAL is a simulator for studying the dynamic behaviour of flow and
congestion control schemes in packet switch data networks. 
It provides users with a way of specifying such networks
and to observe their behavior.
Source code is provided so that interested users can modify the simulator
to their own purposes.
In my experience, anyone who uses a simulator ends up modifying it
anyway, so I expect that users are capable of reading and understanding
source code. This manual, and the programmers manual, are only a brief
sketch of what is available, and what is possible with REAL.
.PP
The simulator takes as input a 
\fIscenario,\fP
which is a description of network topology, protocols, workload
and control parameters.
It produces as output statistics such as the number of packets sent by
each source of data, the queueing delay at each queueing point,
the number of dropped and retransmitted packets and other similar
information.
Scenarios are described using using NetLanguage, a simple
ascii representation of the network.
The network is modeled as a graph, where nodes 
(vertices) represent either
sources or sinks of data, or gateways\(dg. 
.FS
\(dgGateways are used synonymously with routers, bridges and switches
to mean elements that route, buffer and schedule packets.
.FE
The interconnection between the nodes is the topology.
You have to specify some network wide parameters,
the transport protocol (in 
particular, the flow control) and the workload at each source.
Finally, you must specify control parameters such as
the latency and bandwidth of each communication line,
the size of trunk board buffers, packet sizes etc.
.NH 2 
File organization
.PP
The simulator files are in a directory called sim. 
This has four subdirectories - 
docs, sim, src and results.
sim/docs contains online versions of the manuals.
sim/sim has source code for REAL. sim/src has source code
for NEST, the simulation package underlying REAL. 
sim/results stores the results of simulation (it is initially empty).
sim/sim in turn contains some subdirectories described by sim/MANIFEST.
.NH 1 
Sources, Gateways and Sinks
.PP
In REAL, each user is modeled as a source of data regulated
by a flow control protocol.
The combination of a workload and flow control protocol is
implemented by a single C function. Each such function is
executed in parallel by the underlying thread-based simulation package,
and can be thought of as being an independently scheduled
and preemptable entity (see NEST manuals for more details).
.NH 2 
Source types
.PP
There are 11 source types, corresponding to 11 transport
protocol and workload types. 
The sources can be categorized into one of two types:
flow-controlled and non-flow-controlled data sources.
Flow-controlled sources use acknowledgements and timeouts to 
implement a reliable transport layer functionality over a
lossy network.
Non-flow-controlled sources generate data from a known distribution
and do not provide a reliable transport functionality; they model
cross traffic.
.PP
Except for the telnet source, 
the flow controlled sources assume that the users have a finite number
of data packets ready to transfer, and that this data is available
for transmission without any delay.
So, whenever the flow control protocol allows data transmission,
a packet to transmit would always be available.
The telnet source assumes that users generate single packets
with an exponential inter-packet sleep time.
The flow-controlled sources are:
.PP
.B generic
This implements window flow control with fixed size windows.
.PP
.B telnet:
A telnet workload differs from a Poisson workload in two ways
a) the net number of outstanding packets cannot exceed the 
specified window size. 
b) the maximum possible output rate (inverse of the 
interpacket spacing) is limited by the 
bandwidth of the output line from the source (strictly
speaking, a Poisson source can send infinitely fast).
.PP
The next two sources implement TCP flow control with
the window adjustment and other modifications described by
V. Jacobson, M. Karels, P. Karn and C. Partridge [JAC88] [KP91].
.FS
[JAC 88], V. Jacobson, "Congestion Avoidance and Control", Proc.
ACM SigComm '88.

[KP91], P. Karn, C. Partridge,
"Improving Round-Trip Time Estimates in Reliable Transport Protocols",
ACM Trans. on Computer Systems, V9 No. 4, November 1991.
.FE
.PP
.B "jk_tahoe:" 
This implements the flow control in 4.3BSDTahoe. The window size can
increase or decrease - decreases cause the window to be shut down to 1.
Duplicate acks cause retransmission.
.PP
.B "jk_reno:" 
This incorporate further modifications made
to TCP by V. Jacobson (as of Aug '90)\(dg\(dg.
When duplicate acks are seen, 
the window shuts down to half its previous value, not to 1.
.FS
\(dg\(dgThanks to S. McCanne and S. Floyd at Lawrence Berkeley Lab 
for this implementation.
.FE
.PP
.B "dec"
This implements the DEC
congestion bit based window adjusting scheme [RJ 90].
.FS
[RJ 90] K.K. Ramakrishnan and R. Jain, "A Binary Feedback Scheme for
Congestion Avoidance in Computer Networks", ACM Trans. Comp. Sys.
May 1990, V8 N2:158-181.
.FE
.PP
The non-flow-controlled sources send data according to some
parametrized distribution.
The sources in this category are:
.PP
.B "poisson:"
This is a poisson source that sends packets
no faster than the bandwidth of the output line.
Thus, the source is \fIclamped\fP to that speed.
This source is useful in simulating cross traffic, and
in validating queueing models.
.PP
.PP
.B "background:"
is used to create and remove bottlenecks. 
The background source sends data at a chosen rate for
some period of time, 
then is idle for some time, and repeats this cycle.
The fraction of the bottleneck to use is governed by
the average bandwidth parameter in the scenario file.
The source can start at any phase in the cycle, 
can send data with some randomness in the sending rate
Some parameters for this source are set from the file 
background.c.
.PP
.B "controlled_rate:"
This is an on_off source that requests resources using a
SETUP packet. The parameters for the source are
on_time, off_time and peak_rate, described below in the
`REAL parameters' section.
.PP
.B "random_rate:"
This source sends packets so that the sending rate is
no faster than some peak bandwidth, and the average sending rate
is some specified average.
These parameters are described through the scenario file.
Random_rate sources also do call setup.
.PP
.B "mmpp:"
This tries to emulate a Markov-modulated Poisson process data source.
It is not an accurate MMPP model, but adequate for creating
bursty cross traffic.
.NH 2 
Gateway
.PP
The router implements several scheduling disciplines including:
.PP
.B "first come first served"
.PP
.B "fair queueing"
This is described in [DKS 89]. If RESERVATIONS is #defined in 
.FS
[DKS 89] Demers, Keshav, Shenker, "Analysis and Simulation of
a Fair Queueing Algorithm", Proc. SigComm 89 pp1-12, Sep. 1989.
.FE
types.h, then weighted fair queueing is implemented,
and weights will be computed, and buffers reserved, 
so that a source can be assured 
a minimum bandwidth and maximum packet loss rate.
Threshold parameters allow overbooking of resources.
.PP
.B "fqbit:"
is identical to fq, except that it does bit-setting on packets.
Described in [DKS89].
.PP
.B "hrr:"
Implements the HRR discipline [KKK 90].
.FS
[KKK 90] Kalmanek, Kanakia and Keshav, "Rate Controlled Servers
for Very High Speed Networks", Proc. GlobeCom 1990", Dec. 1990.
.FE
.PP
.B "decbit:"
implements DEC bit-setting [RJ 90], and FCFS queueing.
.NH 2 
Sink
.PP
Sink.c implements a sink that receives data from
any source. If the source is a guaranteed service host, the sink only collects
statistics, and does not send an acknowledgment.
Otherwise, it keeps track of the highest in-sequence packet received thus
far, and sends an acknowledgment with that sequence number.
.NH 1 
NetLanguage 4.0
.PP
I now describe NetLanguage, a user-friendly way of 
describing network scenarios.
Refer to sim/lang/example or to the Appendix for an example.
.PP
The header contains identification information for your
own use. It does not affect simulation in any way.
This is followed by Nest parameters.
.PP
.B "passtime:"
How often (in simulation time) the monitor is called. 
I'd recommend 1 second, but you can make it larger if you want
a faster simulator.
An entry of the form a,b is a time value that represents `a' seconds
and `b' microseconds.
.PP
.B "maxnodes:" 
This is the number of nodes in the simulation. 
Set this to the number of nodes in the simulation.
.PP
.B "monitor:"
Selects the monitor (data collection entity) to use. 
You should use `custom_monitor' unless you write your own.
.PP
You now have to set REAL parameters.
Here is a brief explanation of each field.
.PP
.B "ack_size:" 
is the size of the acknowledgment packet (for end-to-end acknowledgement)
in bytes.
.PP
.B "random_seed:" 
This is the seed used to initialize the random number generator.
If this is 0, the seed itseld is chosen at random.
If you use the same random seed, the simulation will be repeated exactly.
.PP
.B "buffer_size:" 
This is the number of bytes of buffer memory that are available
per output line (trunk board). 
.PP
.B "telnet_pkt_size:" 
is the size of a packet sent by the telnet transport protocol in bytes.
.PP
.B "ftp_pkt_size:" 
is the size of a packet sent by a non-telnet source in bytes.
.PP
.B "ftp_window, telnet_window:" 
are the maximum sizes of ftp and
telnet windows. (These are not the same as the Jacobson congestion 
windows, but are their upper bounds.) The window size should 
not exceed the MAX_WINDOW_SIZE parameter in sim/sim/config.h.
.PP
.B "decongestion_mechanism:" 
This selects which packet to drop
when a queueing buffer is full.
If the paramter is 0 or 1, then the packet to drop depends on
the scheduling discipline.
For Fair Queueing or HRR, first, the conversation with the most bytes in
the queue is selected.
Then, 
0 drops the packet at the head of the queue and
1 drops the packet at the tail.
If the scheduling discipline is FCFS or DECbit, 
both 0 and 1 will drop the incoming packet.
When this parameter is set to 2, the simulator selects a packet 
at random amongst all the queued packets and drops it.
.PP
.B "policy:" 
This is used to select the scheduling policy at the gateways. 
FCFS is 1, Fair Queueing [DKS 89] is 2, 3 is
FCFS + bit setting (DEC bit)[RJ 90], 
4 is Fair Queueing + bit setting,
and 5 is Hierarchical Round Robin(HRR [KKK 90]).
.PP
.B "router,"
.B "real_number:" 
Used in distributed simulation. See the description on distrib/README.
.PP
.B "end_simulation:"
This is the simulation time at which to end the simulation run.
.PP
.B "print_interval:"
Number of seconds of simulation time after which to print a report. 
All statistics, except min and max values, are refreshed
every print_interval seconds.
.PP
.B "hrr_levels:"
value for nb and b, as defined in [KKK 90]. If all the servers use the
same value, it can be described here, else per node.
.PP
.B "scale_factor:"
scales simulation time, so, for example, if scale_factor is 1000,
one simulated second corresponds to one millisecond. This
is mainly used to set initial values for hard-coded constants
in transport layer protocols (such as the initial estimate of
the round trip time).
.PP
Next, node and edge parameters are defined. 
To make it easier to declare the nodes,
the first declaration in the block is the default declaration.
The default defines the parameters of each node
unless they are overridden by an explicit redeclaration
inside some node.
You can skip any entries in the
defaults declaration if the simulation does not use them, 
and you may skip as many
entries in the node declaration as you wish.
.PP
The function field is the name of the
C function that should run on that node.
`dest' is the ID of the sink to which data should be sent.
`start_time' is the simulation time at which the node becomes active. 
You should set plot to true if you want the node
function at that node to generate plots (described later).
The other values are self explanatory, and are commented in
sim/lang/example.
.PP
Edges are defined by source, destination, bandwidth and
the latency in communication. 
The endpoints of an edge are specified as source->destination,
and all edges are bidirectional.
The bandwidth is specified in bits per second, and the delay in 
microseconds.
.NH 2 
A note on usage
.PP
Usually, you will not need to change the nest parameter and
function declaration sections of the language file.
In most cases, you should just copy these from the
file sim/lang/example.l.
It is necessary to give a .l extension to a language file.
.PP
It is often useful to declare default values in a file,
and to include them in the language file, particularly
when several simulation runs are batched.
One trick is to create a .s (for small) file that
#includes some default files.
The `asm' script in sim/sim/scenarios will assemble
a .l file from a .s file, and may be of use.
.PP
Finally, the results of the simulation will be placed in 
a subdirectory of the current working directory.
.NH 1 
Running the simulation
.PP
Make sure that the FUNC_TABLE file exists in the current directory.
Then, take a deep breath, and type
.ce10

simulate <language_file.l>

.ce0
.NH 1 
Simulation output
.PP
The simulator produces two kinds of outputs
a) time ticks and c) a simulation report.
.PP
Time ticks are printed out every `print_interval' 
seconds of simulated time. 
They are just to show you that something is going on.
.PP
Reports are generated every `print_interval' seconds. 
They are appended to a result file.
The report file resides in cwd 
concatenated with the input filename without the .l.
The report file is named `dump'.
For example, if your input is foo.l, 
then the report will be in ./foo/dump.
There are a couple of options: you can type in the input file 
without the .l if you wish, and REAL will complete the
file name for you (of course, the file itself must exist
with the .l).
Also, if the input file is supplied through indirection,
then the standard output file defaults to `../results/dump'.
.PP
The report is in the form of table and will look like this:

.KF
.nf
\fC
--------------------------------------------------------------------------------------
Time   #  Type           G/w   Xmit  Q'ing(min,ave,max)[#]  Drops Retx RTT(min,ave,max)
--------------------------------------------------------------------------------------

300    1   Generic        5(FC) 1434  0.29 0.93 1.19 [4]  0     1     0.00 1.04 1.30
       2   Generic        5(FC) 1436  0.10 0.93 1.19 [5]  0     0     0.00 1.04 1.30
       3   Telnet         5(FC) 63    0.20 0.99 1.25 [0]  0     1     0.00 1.10 1.36
       4   Telnet         5(FC) 67    0.00 0.97 1.16 [0]  0     1     0.00 1.08 1.27

600    1   Generic        5(FC) 1443  0.89 0.93 1.19 [5]  0     0     1.00 1.04 1.30
       2   Generic        5(FC) 1441  0.89 0.93 1.19 [4]  0     0     1.00 1.04 1.30
       3   Telnet         5(FC) 55    0.89 0.97 1.12 [0]  0     0     1.00 1.08 1.23
       4   Telnet         5(FC) 61    0.89 0.98 1.18 [0]  0     0     1.00 1.09 1.29


Summary

Node G/w T'put(mean,var) Q'ing(mean,var) RTT(mean,var)   Drops(mean,var) Retxs(mean, var)

1    5   (1437.17 4.32  ) (0.93   0.00  ) (1.04   0.00  ) (0.00   0.00  ) (0.00 0.00  )

2    5   (1437.33 4.37  ) (0.93   0.00  ) (1.04   0.00  ) (0.00   0.00  ) (0.00 0.00  )

3    5   (58.17  3.62  ) (0.98   0.01  ) (1.09   0.01  ) (0.00   0.00  ) (0.00 0.00  )

4    5   (67.33  8.42  ) (0.98   0.00  ) (1.10   0.02  ) (0.00   0.00  ) (0.17 0.37  )

.\fP
.fi
.KE
Every print_interval, a report of simulation statistics is printed.
The first entry is the time (in this example, 300 seconds).
For each source, the following data is printed:
.PP
.B "#:"
This is the node ID that of the node.
.PP
.B "Type:"
This is the name of the node function running on the node.
.PP
.B "G/w:"
This is gateway number for which the delays are reported.
If the source sends packets through multiple gateways, then 
the report will print out data at each of the intermediate
gateways.
The gateway number is followed by the scheduling policy at the router.
FC is for a router obeying FCFS, FQ, Fair Queueing, 
FB, Fair Queueing with Bit Setting, DB, DEC Bit setting,
HR Hierarchical Round Robin.
.PP
.B "Xmit:"
This is the number of packets transmitted from that gateway that originated
from that source (including retransmissions) during the previous
print_interval seconds of simulation time.
.PP
.B "Q'ing:"
The minimum, average and maximum
queueing delays, are printed, followed by (in square brackets) a 
snapshot of the number of packets queued in the per-conversation
queue for that source at the time the report was printed.
The queueing delay is computed as the time a packet 
had to wait in the queue from the time that its last bit 
arrived at the gateway 
to the time that the first bit in that packet was placed
on the output trunk.
.PP
.B "Drops:"
This is the number of packets from that source
that were dropped by intermediate gateways.
.PP
.B "Retx.:"
This is number of packets that the source retransmitted.
.PP
.B "Rtt:"
This is the min, average and max round trip time that packets from that
source took from the time that the first bit of the packet was placed
on the line to the time that the last bit of the acknowledgment arrived.
In case of JK sources, it is computed only for packets that
have never been retransmitted.
.PP
In order to avoid corrupting simulation results with
data from transients that happen at start up, 
simulation statistics are flushed after each dump.
In other words, the report presents statistics for the
last print_interval seconds of simulation only.
.PP
Note that to compute effective throughput you must subtract the number
of retransmissions from the number of packets sent by the gateway.
The effective load is the number of packets transmitted by the gateway 
plus the number of packets it dropped.
.PP
If print_interval is long enough, 
the mean value printed out at the end of the 
interval will be a `large enough' sample of
the path of a variable, and so can be thought of as a batch mean. 
From elementary statistics, we know that 
the batch mean will be normally distributed.
REAL computes both the mean of batch means, 
and the standard deviation of these means. 
Using the normal assumption, 99% confidence will be achieved
at 3 SD on either side of the mean of means.
.PP
At the end of the simulation, 
the mean of batch means and SD of batch means is printed.
This is headed by a line marked `Summary'.
Then, for each value in the throughput, the queueing delay at each gateway, 
the number of packet losses from each source in each interval
and the number of retransmissions in each interval are 
the mean of means and SD of means is printed.
To avoid start-up transients from affecting the
variance value, the parameter NSKIP #defined in table.c
selects how many of the initial time periods to skip.
The default value is 1.
.NH 1 
Plotting simulation variables
.PP
The node functions in REAL have been instrumented to produce traces of 
certain variables as the simulation proceeds. 
I'll describe the nature of these traces, and then how you can plot them
out on the screen or on a printer. Finally, I'll describe how you can
trace additional variables.
.PP
The plotting function is trivially implemented. 
Node functions are coded to print out the simulation variable whenever 
the variable could change, along with the current simulation time.
For example, TCP sources print out the size of the congestion control
window whenever the congestion window could change.
These (time, value) tuples can then be post processed to derive
other statistics, such as average values, or can be plotted out.
.PP
To plot the time vs. value graphs for these variables, 
you can use a standard package such as ggraph or xgraph.
The files that plot creates are of the form CCCxx, where CCC is an 
identification tag and xx is the node ID of the node
which is producing the plot.
The files will appear in the same directory as the report file
(see the description above).
The currently available files are 
.PP
.B "rtt:"
Contains the estimated round trip times.
.PP
.B "tao:"
Contains the \fIactual\fP
round trip times.
.PP
.B "win:"
Contains the congestion control window size.
.PP
.B "e2e:"
This is the sequence of end-to-end delays seen by packets
from each source during the course of the simulation. 
This can be post-processed to compute the frequency histogram of the
end-to-end delay etc.
.B "seq:"
The sequence number of the packet transmitted at that time.
.sp1
.PP
In each case, the X axis is time in seconds, and the Y axis is the data value.
.PP
To add your own instrumentation, you have to edit the node functions in sim/sim.
Choose the variable that you want to plot, and at the appropriate
point(s), insert the function call `make_plot(filename, variable);'.
For example, to plot the window size, the command I use is
`make_plot("win", cur_window);'.
To plot real numbers, use `make_flt_plot()'.
Recompile the simulator (`make'), 
and the plot files will be produced automatically
for each node that has the plot option set.
Use `make clean' to remove old plot files.
.NH 1 
A Note on REAL Configuration
.PP
Note that some hard limits on simulation size are defined in 
the file config.h. If you plan to do large simulations, 
you should change this file and recompile the binary.
.PP
If you would like to do distributed simulation, 
something that is explained in the REAL technical report and in the
programmer's manual,
you should #define DISTRIBUTE.
(The distributed simulation facility is currently 
unsupported, but you are welcome to try it at your own risk.)
If you want to weighted Fair Queueing, 
you should #define RESERVATIONS in sim/types.h.
.PP
You should run the simulator binary in the sim/sim directory.
If you do move it out of there, redefine FUNC_FILE in 
the makefile.
.NH 1 
Acknowledgements
.PP
S. McCanne at Lawrence Berkeley Lab contributed some code for
jk_reno.c and plotting.c, and reorganized the code in jk_tahoe.c. 
I have reimplemented some features based on his additions to REAL.
S. Jamin at USC added the ability to specify result files.
K.K. Ramakrishnan at DEC helped with the testing of DEC sources.
R. Sethi at Bell Labs suggested the use of defaults in NetLanguage.
D. Ferrari at UCB provided advise on statistics and other features. 
Last but not least, 
S. Shenker at Xerox PARC was the main instigator of REAL, and has provided
invaluable advice over the years.
.bp
.sp2
.(c
.B "APPENDIX : NetLanguage sample"
.)c
This is a sample description in NetLanguage that 
is in sim/lang/lang.inp.

.nf
# This is a sample input file for the simulator
# it can also be found as lang/example.l
# To see if your simulator is OK, run it on this example as 
# simulate example.l

# the header section is optional

header					
{
network: example;
version: 0;
}

# nest_params section is optional. Default passtime is 1 second
# default maxnodes is MAXNODES in config.h and default monitor is
# custom_monitor in kernel/monitor.c

nest_params {
	passtime = 1,0;
	maxnodes = 30;
	monitor = custom_monitor;
}

# real_params section is also optional. Default definitions are in
# lang/lang.yacc

real_params {
	inter_pkt_delay = 5.0;
#                                       seconds (optional - not used)
	random_seed = 23;
	buffer_size = 440000;
#					bytes 	(different from REAL3.0)
	ftp_pkt_size = 500;
#                                       bytes
	telnet_pkt_size = 40;
#					bytes
	ack_size = 40;
#                                       bytes
	ftp_window = 600;
#                                       packets
	telnet_window = 400;
#                                       packets
	decongestion_mechanism = 1;
#					1 = drop from tail
#					2 = drop from head
#					3 = drop from random position
	real_number  = 0;
#					used for distributed REAL
	router_node = 0;
#					used for distributed REAL
	util_time = 1.0;
#					time interval over which to 
#					measure router utilization (secs)
	end_simulation = 100;
#					simulation end time in seconds
	print_interval = 10;
#					print and flush statistics every 
#					print_interval seconds
	scale_factor = 1;
#					time scale factor; if scale factor =
#					1000, 1 simulated second = 1ms realtime
}

# this section is only for backwards compatibility with REAL 3.0
# it is optional in REAL 4.0

node_functions {
#					This section is for documentation only:
#				  	and you can skip it if you want.
#					if you do use it, the name must match
#					the function name in the source file.
	poisson;
	router;
	sink;
}

nodes{ 
# 					first define the default node type
#					all fields are optional
	default{
		function = poisson;
		dest = 4;
#					destination node to send to 
		start_time = 0,10000;
#					start node at seconds, microseconds
		plot = false ;
#					set to true if you want to plot data
		num_pkts = 1000;
#					number of packets to send 
		on_time = 0;
#					for on-off style sources:
#					on_time is the time from the start
#					of the sending of the first
# 					packet in the on period to the start 
#					of sending of the last pkt.

		off_time = 40000;
#					off_time is time between start
#					of last pkt. in prev burst
#					to start of first pkt in next burst
		peak = 20000.0;
#					for controlled sources: this is the
#					declared peak rate in bits/sec
		average = 5000.0;
#					declared average rate in bits/sec
		interval = 1000000;
#					interval over which average is computed
#					(in microseconds)
		jitter = 20000000;
#					desired max delay jitter from the 
#					network (in microseconds)
		hrr_levels = { 10:0 ; } ;
#					for a hrr router, the b and nb values
#					at each level. Ignored for a source.
		sch_policy = 2;	
#					for router nodes 
#					1 = fcfs
#					2 = fair queueing
#					3 = decbit
#					4 = fqbit
#					5 = hrr
		}

#					for each source, just need to specify
#					what is different compared to the default
	node 1 { function = poisson; plot = true; num_pkts = 4000; dest = 3;}
	node 2 { function = router; plot = true;}
	node 3 { function = sink;}

}

edges{
	default {
		bandwidth = 2000000;
#					in bits/sec
		latency = 10000;
#					in microseconds
		}
	{ 1 -> 2;}
	{ 2 -> 3; bandwidth = 400000; latency = 20000;}
}
.fi
