/*
 * hash.c 
 * hash and inverse hash functions
 */


#include "real.h"

char            used[MAX_NODES + 1][MAX_CONVERSATIONS];
int             source[MAX_NODES + 1][MAX_CONVERSATIONS];
int             dest[MAX_NODES + 1][MAX_CONVERSATIONS];

/* since the number of conversations is small, use just a linear scan */

int
hash (node, so, de)
    ident           node;
    int             so, de;
{
    int             conv_id;
    char            h_flag = 0;

    /* conv_id = (7 * so + 11 * de) % MAX_CONVERSATIONS; */
    /* see if its already there */
    for (conv_id = 0; conv_id < MAX_CONVERSATIONS && h_flag is 0;)
	if (used[node][conv_id] &&
	    source[node][conv_id] is so &&
	    dest[node][conv_id] is de)
	    h_flag = 1;
	else
	    conv_id++;
    if (conv_id >= MAX_CONVERSATIONS)
	/* not there */
    {
	/* look for space to put it in */
	for (conv_id = 0; (conv_id < MAX_CONVERSATIONS)
	     && used[node][conv_id]; conv_id++);
	if (conv_id >= MAX_CONVERSATIONS)
	    /* no space */
	    pr_error ("No more space in hash table ");
    }
    used[node][conv_id] = 1;
    source[node][conv_id] = so;
    dest[node][conv_id] = de;
    return conv_id;
}

invert (node, conv_id, so, de)
    ident           node;
    int             conv_id, *so, *de;
{
    if (!used[node][conv_id])
	return -1;
    *so = source[node][conv_id];
    *de = dest[node][conv_id];
    return 0;
}

free_hash_entry (node, conv_id)
    ident           node;
    int             conv_id;
{
    used[node][conv_id] = 0;
}

/*
testing ...

main()
{
int conv;
int so, de;

conv = hash (10, 110, 20);
printf ("%d \n", conv);
invert (10, conv, &so, &de);
printf ("%d, %d \n",so , de);
conv = hash (10, 4,14);
printf ("%d \n", conv);
invert (10, conv, &so, &de);
printf ("%d, %d \n",so , de);
free_hash_entry (10, 0);
invert (10, 1, &so, &de);
printf ("%d, %d \n",so , de);
conv = hash (10, 110, 20);
printf ("%d \n", conv);
invert (10, 0, &so, &de);
printf ("%d, %d \n",so , de);
}


*/
