#include "real.h"
#define extern
#include "../../src/message.h"
#undef extern
#include "../../src/network.h"
#include "../../src/simulate.h"
#include "../../src/process.h"

/*
 * set a timer for delay seconds. At this time, the pkt is sent to the
 * sending node 
 */

set_timer (delay, pkt)
    float	  delay;
    PKT_PTR         pkt;
{
    timev           now, tv_delay;

    if(delay < 0){
	pr_error("set_timer given negative value for delay !\n");
    }
	
    now = runtime ();
    if (pkt is NULL)
	pr_error ("set timer got a null pkt");

    tv_delay.tv_sec = (long) delay;
    tv_delay.tv_usec = (long)( 1.0e6 * (delay - (float)((long) delay)));

    tv_delay.tv_usec += now.tv_usec - Graph->header->timenow.tv_usec;
    tv_delay.tv_sec += now.tv_sec - Graph->header->timenow.tv_sec;

    if(tv_delay.tv_usec > MILLION) {
	tv_delay.tv_sec ++;
	tv_delay.tv_usec -= MILLION;
    }

/*
    delay += MILLION * (now.tv_sec - Graph->header->timenow.tv_sec)
	+ (now.tv_usec - Graph->header->timenow.tv_usec);
*/
    own_reliable (_current_node, _current_node, _current_node, 0, pkt, tv_delay, Graph->header->timenow);
}

/* VARARGS 6 */

int
own_reliable (a_node, sender, dest, key, dataptr, delay, now)
    ident           a_node,	/* actual destination node (never 0) */
                    sender,	/* sender (usually _current_node) */
                    dest;	/* destination node (0 if all nbrs) */
    long            key;	/* message code (user defined int) */
    char           *dataptr;	/* pointer to message data (on heap) */
    timev           delay;	/* time at which message will arrive */
    timev           now;	/* time at the monitor */
{
    messlist      **mlist, *newitem;
    message        *newmsg;


    if (a_node is 0 or a_node > Nodes or dataptr is 0)
	/* a_node must be a valid node */
	pr_error ("own_reliable got a bad packet ");

    /* create "message_unit" and "message_list" structures */

    newitem = new (messlist);
    newitem->msg = newmsg = new (message);

    /* set "message_unit" fields */

    newmsg->key = key;
    newmsg->dataptr = dataptr;
    newmsg->sender = sender;
    newmsg->dest = dest;

    /* update sent and arrived time and blocked status of new message */

    _myreschedule (a_node, delay, &(newmsg->sent), &(newmsg->arrived), now);

    /* insert message into message list for this node */

    mlist = &_message_table[a_node];
    while ((*mlist) isnt nil and
	   not time_after ((*mlist)->msg->arrived, newmsg->arrived))

	advancn (mlist);	/* search for later message */

    insn (newitem, mlist);	/* insert new message at this point */

    return (1);			/* currently, this always succeeds */
}

_myreschedule (nodeid, later, nowtime_ptr, arrivedtime_ptr, now)
    ident           nodeid;
    timev           later;
    timev          *nowtime_ptr, *arrivedtime_ptr, now;

/* patch *nowtime_ptr to now */
{
    *nowtime_ptr = now;
    *arrivedtime_ptr = *nowtime_ptr;

    /* arrived time is the current runtime plus the delay */
/*
    this code was when delays were long quantities 

    arrivedtime_ptr->tv_sec +=
	(arrivedtime_ptr->tv_usec += later) / MILLION;
    arrivedtime_ptr->tv_usec %= MILLION;
*/
    arrivedtime_ptr->tv_usec += later.tv_usec;
    arrivedtime_ptr->tv_sec += later.tv_sec;

    if(arrivedtime_ptr->tv_usec > MILLION) {
        arrivedtime_ptr->tv_sec ++;
        arrivedtime_ptr->tv_usec -= MILLION;
    }
    if (_proc_table[nodeid].whyblocked is RECEIVE
	or _proc_table[nodeid].whyblocked is WAIT)

	/* deleted RECV */
    {
	if (time_iszero (_proc_table[nodeid].unblocked))
	    /* remove from blocked queue */
	    _remove_sched (nodeid, BlockedQueue);
	else
	{			/* arrive before unblocked time */
	    if (time_before (*arrivedtime_ptr,
			     _proc_table[nodeid].unblocked))
	    {

		/* if unblocked less than sync_time, remove from
		 * waitqueue, otherwise, remove from newpass waitqueue */

		if (time_before (_proc_table[nodeid].unblocked, _sync_time))
		    _remove_sched (nodeid, WaitQueue);
		else
		    _remove_sched (nodeid, NewpassWaitqueue);
	    } else		/* no rescheduling necessary */
		return;
	}
	_proc_table[nodeid].unblocked =
	    when (time_after (*arrivedtime_ptr, _proc_table[nodeid].runtime))
	    (
	     then (*arrivedtime_ptr)
	     otherwise (_proc_table[nodeid].runtime)
	    );

	_insert_sched (nodeid);
    }
}
