/*
 * This file is supposed to be the same as PP with some lines ifdeffed out.
 * The function defined here, pp_reg, provides pp style regulation, but does
 * not do any flow control - only good timeouts and retransmissions.
 * This is done by disabling TIMERs
 */

/* Revision history */

/* 5/12/94: fixed B probing, and probing parameters */
/* 5/16/94: sink now sends ack_for as well as seq_no - changed retx code */
/* 5/18/94: added PP_TRANS changes to test */
/* 5/20/94: debugged PP_TRANS - num_out, use of last_sent instead of pkts_sent
	    added old_last_ack heuristic for lost retxs */

extern int      PP_PRINT;
extern int      BIG_CTH_DEBUG;

#include "../kernel/real.h"
#define PP_TRANS
#define NO_FLOW_CONTROL

extern float    compute_alpha();

#undef	FILTER		     	/* if defined, then probes are filtered */
#define ATTACK_UP       1.5	/* attack rate if bottleneck less full */
#define ATTACK_DOWN     0.8	/* attack rate if bottleneck less full */
#define TIMEOUT_MULTIPLIER 1.5	/* this times RTT est. is timeout value */
#define INITIAL_TIMEOUT  1.0    /* initial timeout in seconds */
#define INC_FACTOR 	0.2	/* factor controlling send rate increase */
#define MS_ALPHA        0.75	/* exp. av. const. for mean_send_rate */

#define B_DEC_FACTOR	0.75	/* multiplicative decrease factor */
#define B_ADD_INC_FACTOR 2.0 	/* additive increase for setpoint probe */
#define B_INITIAL	5	/* initial value of B */
#define B_MIN		2.0	/* smallest value */
#define B_COUNT		2	/* how many rtts to wait before changing B */


#define ceil(x) (int)((x) + 1)
#define MWS MAX_WINDOW_SIZE
extern int         num_retx[MAX_NODES+1][MWS];		/* number of times seq # retx */
extern int	    recvd_ok[MAX_NODES+1][MWS];		/* this seq. no recvd ok */

#ifdef NO_FLOW_CONTROL
pp_reg()
#else
pp()
#endif
{
    PKT_PTR     deq_pkt; 		/* pkt dequeued from tx queue */
    PKT_PTR	pkt;			/* pkt under consideration */
    PKT_PTR	retx_pkt;  		/* pkt to be retransmitted */
    int         i, num;			/* scratch variable */
    int	    	node;			/* ID of current node */
    int		seq_no = 0; 		/* sequence number counter */
    int	  	last_sent = -1;		/* highest seq no sent so far */
    int         num_outstanding = 0;	/* number of packets outstanding */
    int         start_up = 1;		/* flag indicating startup */
    int		line_busy = 0; 		/* flag indicating output line byust */
    int		tick;			/* how many packets to send to probe */
    int		num_timeouts = 0;	/* # timeouts seen so far */
    ident       destn, sender, sink;	/* scratch variables */
    long        key;			/* scratch variables */
    timev       now;			/* time now */
    timev	diff; 			/* scratch */
    timev	time_of_last_ack;	/* time when last ack recvd */
    float       tao;			/* scratch */
    float	timeout; 		/* timeout value */
    float	alpha;			/* exp. av. const for rate estimate */
    float	nbhat = 0;		/* estimate of # pkts in bottlneck */
    float       se;			/* bottleneck rate estimator ^-1 */
    float	re;			/* estimator of RTT */
    float	reclean;		/* clean estimate of prop. delay */
    float	send_rate;		/* computed sending rate */
    float	old_send_rate = 0.0;	/* old sending rate */
    float	mean_send_rate = 0.0;	/* smoothed sending tate */
    float	inter_ack;		/* current value */
    int         last_ack = -1; 		/* last ack seen so far */
    int         old_last_ack = -1; 	/* last ack seen so far */
    int 	first_of_pair;		/* seq # of first ack of pair */
    int	 	transmission_id = 0;	/* id to simulate single timer */
    int         dup_ack_retx = 0;	/* # retx. pkt due to dup ack */
    int 	total_dup_acks = 0; 	/* # pkts retx.ed so far from dupacks */
    float    	B = B_INITIAL;		/* setpoint */
    int 	rtt_seq = 0;		/* if ack > rtt_seq, >= 1 RTT has elapsed */
    int		rtt_count = 0;		/* how many rtts have gone by */
    char 	no_more_data = 0;	/* true if app says theres no more data */
    int		app_node = 0;		/* node number of application layer */

    stop_time();
    node = get_node_id();
    abs_advance(node_start_time[node]);
    sink = assigned_sink[node];
    printf("pp source: %d ", get_node_id());
    printf("sending %d packets to sink %d\n", num_pkts[node], sink);
    source_node_type[node] = PP_SOURCE;
    for (i = 0; i < MAX_WINDOW_SIZE; i++){
	num_retx[node][i] = 0;
	recvd_ok[node][i] = 0;
    }
    last_ack = -1;
    time_of_last_ack = runtime();
    timeout = INITIAL_TIMEOUT * scale_factor;
#ifdef NO_FLOW_CONTROL
    tick = 1;
#else
    tick = 2;
#endif
    line_busy = 0;
    goto test;

    for (ever)
    {

recv:
	sender = recvm(&destn, &key, &pkt);
	now = runtime();
	if (BIG_CTH_DEBUG)
	    printf("\n%f: node %d recvd. packet seq %d, resent %d last_ack %d  type %s, decbit %d\n", 
		make_float(now), node, pkt->seq_no, pkt->resent, last_ack, 
		type_to_str(pkt->type), pkt->decbit);

	switch (pkt->type)
	{

	case ACK:
	    if(PP_PRINT)
	    	make_plot("ack", pkt->seq_no);

	    if(PP_PRINT)
	    	make_plot("num_out", num_outstanding+1);

	    num_outstanding--;
	    /* since at least one packet reached the sink */

	    if(PP_PRINT)
	    	make_plot("num_out", num_outstanding);

	    /* once in B_COUNT  RTTs, we will increase B */

	    if(pkt->id >= rtt_seq){
		rtt_count ++;
		rtt_seq = transmission_id;
	    }  

	    /* every 2 rtts, check to see if we have made progress,
	     * if not, retx the last ack, since it might have been lost again 
	     */

	    if(rtt_count is 2) {
		if(last_ack is old_last_ack) {
		    make_pkt(retx_pkt);
		    retx_pkt->seq_no = last_ack + 1;
		    retx_pkt->resent = 1;
		    num_retx[node][retx_pkt->seq_no % MWS] ++;
	            if(PP_PRINT)
		        make_plot("num_out", num_outstanding);
		    num_outstanding --;		
	            if(PP_PRINT)
		        make_plot("num_out", num_outstanding);
		    enq_high(node, retx_pkt);
		    dup_ack_retx ++;
		}
		old_last_ack = last_ack;
	    }
		    
	    if (rtt_count is B_COUNT)
	    {
		rtt_count = 0;

		if(PP_PRINT)
			make_flt_plot("B", B);

		B += B_ADD_INC_FACTOR;

		if(PP_PRINT)
			make_flt_plot("B", B);
	    }

	    recvd_ok[node][(pkt->seq_no + pkt->last_recvd_offset) % MWS] = 1;

	    if (pkt->seq_no is last_ack) {
		total_dup_acks++;
	        for(i = 1; i < pkt->last_recvd_offset; i++) {
		    if (((num_retx[node][(last_ack + i) % MWS]) is 0)
			and 
			(not (recvd_ok[node][(last_ack + i) % MWS]))){
	    	        if(BIG_CTH_DEBUG)
		            printf("%f : node %d: dup_ack seq_no %d enqueued\n", 
		    	        make_float(now), node, (last_ack + i));
		        make_pkt(retx_pkt);
		        retx_pkt->seq_no = last_ack + i;
		        retx_pkt->resent = 1;
			num_retx[node][retx_pkt->seq_no % MWS] ++;
	                if(PP_PRINT)
		            make_plot("num_out", num_outstanding);
			num_outstanding --;		
	                if(PP_PRINT)
		            make_plot("num_out", num_outstanding);
		        enq_high(node, retx_pkt);
		        dup_ack_retx ++;
	            }
		}
		if (total_dup_acks is 1){
	            if(PP_PRINT)
		        make_flt_plot("B", B);
		    B = (B*B_DEC_FACTOR > B_MIN)? B*B_DEC_FACTOR:B_MIN;
	            if(PP_PRINT)
		        make_flt_plot("B", B);
	        }
	    }

	    if (pkt->seq_no > last_ack)
	    {
		for (i= last_ack; i <= pkt->seq_no; i++)
		     num_retx[node][i % MWS] = 0;
		last_ack = pkt->seq_no;
		total_dup_acks = 0;
	    }

	    /* compute round trip time */

	    diff = time_minus(now, pkt->gen_time);
	    tao = make_float(diff);
	    make_entry(tao, &rt_time[node]);

	    switch (pkt->decbit)
	    {
	    case 0:	     /* first in burst (pair) */
		if (start_up)
		    reclean = tao;
		if (BIG_CTH_DEBUG)
		    printf("reclean %f\n", reclean);
		time_of_last_ack = now;
		first_of_pair = pkt->id;
		free(pkt);
		break;

	    case 1:	     /* second in burst */
		if ((pkt->id is first_of_pair + 1))
		{
		    first_of_pair = -2;
		    /* just to be safe */

		    diff = time_minus(now, time_of_last_ack);
		    if (PP_PRINT)
		        make_flt_plot("inter_ack", inter_ack);
		    inter_ack = make_float(diff);
		    if (PP_PRINT)
		        make_flt_plot("inter_ack", inter_ack);

		    if (!start_up)
		    {
		        if (BIG_CTH_DEBUG)
		    	    make_flt_plot("alpha", alpha);
			alpha = compute_alpha(se, inter_ack);
		        if (BIG_CTH_DEBUG)
			    make_flt_plot("alpha", alpha);
			se = alpha * se + (1 - alpha) * inter_ack;
		        if (PP_PRINT)
			    make_flt_plot("nbhat", nbhat);
			nbhat = num_outstanding - (reclean / se);
			nbhat = (nbhat > 0) ? nbhat : 0.0;
			nbhat = (nbhat > ftp_window) ? ftp_window : nbhat;
		        if (PP_PRINT)
			    make_flt_plot("nbhat", nbhat);


			/*
			 * if we are pushing queue up, can be more
			 * conservative than if we are pushing queue down
			 */

			if (nbhat <= B)
			    re = ATTACK_UP * (reclean + nbhat * se);
			else
			    re = ATTACK_DOWN * (reclean + nbhat * se);


			/* control law from thesis page 67 */

			send_rate = (B - nbhat)/re + 1/se;

			/*
			 * special case: if nbhat is very small (<2), then
			 * it is likely that the probe values are wrong.
			 * In this case, it is better to be conservative
			 */

			if (nbhat <= 2.0 and (send_rate > mean_send_rate))
			{
			    send_rate = (mean_send_rate +
				(send_rate - mean_send_rate) * INC_FACTOR);
			}
			mean_send_rate = mean_send_rate * MS_ALPHA +
			    (send_rate) * (1 - MS_ALPHA);

			/* special case: if we have overflowed bottleneck by 
			   a large amount, and so send rate is too low, set
			    timer to the time that the queue would be drained */

			if ((nbhat > 3*B) and (send_rate < 2.0 /((nbhat-B)*se)))
			    send_rate = 2.0/((nbhat-B)*se);

				/* time taken to drain queue to B */

		        if (PP_PRINT) {
			    /* make_flt_plot("mean_send_rate", mean_send_rate);*/
				
			}

			if(BIG_CTH_DEBUG)
			    printf("node %d: nbhat %f , num_out %d, re %f, inter_ack %f, se %f, send_rate %f alpha %f\n", node, nbhat, num_outstanding, re, inter_ack, se, send_rate, alpha);
			timeout = TIMEOUT_MULTIPLIER * re;
			free(pkt);
		    } else
		    {	    /* got first pair of acks at time RTT */
			re = tao;
			se = inter_ack;
			start_up = 0;
			send_rate = 1/se;

			send_rate = (mean_send_rate +
				(send_rate - mean_send_rate) * INC_FACTOR);
			mean_send_rate = mean_send_rate * MS_ALPHA +
			    			(send_rate) * (1 - MS_ALPHA);
			if(BIG_CTH_DEBUG)
			    printf("node %d: startup inter_ack %f, se %f, send_rate %f \n", node, inter_ack, se, send_rate);

#ifndef NO_FLOW_CONTROL
			/* convert to first tick; compute tick size */
			pkt->type = TIMER;
			set_timer((float) (2.0/send_rate), pkt);
			tick = 2;
#endif
			goto test;
		    }	     /* start up */
		} else {
		    first_of_pair = -2;
		    time_of_last_ack = time_zero;
		    free(pkt);
		}
		break;
	    default:
		free(pkt);
		break;
	    }
	    goto test;
	case INT:
	    line_busy = 0;
	    free(pkt);
	    goto test;
#ifndef NO_FLOW_CONTROL
	case TIMER:
	    if(BIG_CTH_DEBUG)
	        printf("setting tick of %d\n", (int) (2 * MILLION /send_rate));
	    if(last_sent < num_pkts[node]){
	        set_timer((float) (2.0/ send_rate), pkt);
	    }

	    if(PP_PRINT) {
	        make_flt_plot("send_rate", old_send_rate);
	        make_flt_plot("send_rate", send_rate);
	        old_send_rate = send_rate;
	    }

	    tick = 2;
	    goto test;
#endif
	case TIMEOUT:
	    /*
	     * put all unacked packets in the retx. queue. If an ack
	     * arrives, it will clean out the extra pkts
	     */
	    if (pkt->id is (transmission_id - 1))
		/* simulate single timer */
	    {
		if(PP_PRINT){
		    make_plot("timeout", 0);
		    make_plot("timeout", 1500);
		    make_plot("timeout", 0);
		}

	        free(pkt);
	 	if(BIG_CTH_DEBUG)
		    printf("%f : node %d: timeout on seq_no %d\n",
			   make_float(now), node, pkt->seq_no);

		if (PP_PRINT)
	    	    make_flt_plot("B", B);
		B =  B_INITIAL;
		if (PP_PRINT)
	    	    make_flt_plot("B", B);

		for (i = last_ack + 1; i <= last_sent; i++)
		{
		    if(!recvd_ok[node][i % MWS]) {
		              /* dont change value of timeout - no backoff */
		        make_pkt(retx_pkt);
		        retx_pkt->seq_no = i;
		        retx_pkt->resent = 1;
		        enq_high(node, retx_pkt);
		        num_timeouts++;
		    }
		}
	    } 
	    else
		free(pkt);

	    goto test;

#ifdef PP_TRANS
	case DATA:
	    app_node = pkt->source;
            pkt->dest = sink;
            pkt->resent = 0;
            pkt->seq_no = seq_no++;
            pkt->source = node;
	    enq(node, pkt);
	    goto test;

	case NO_MORE_DATA:
	    no_more_data = 1;
	    free(pkt);
	    goto test;

#endif
	default:
	    pr_error("Node %d: pp source recvd. an unknown pkt ", node);
	}

test:
	num = num_in_q(node);
	if(PP_PRINT)
	    make_plot("num_q", num);
	if((not no_more_data) and (num < tick)) {
#ifdef PP_TRANS
	    if(app_node isnt 0 and num is 0 and seq_no <= last_ack + 1){
  	        make_pkt(pkt);
	        pkt->type = TX_Q_EMPTY;
	        sendm(app_node, 0, pkt);
	    }
	    goto recv;
	}
#else
	    make_pkt(pkt);
	    pkt->dest = sink;
	    pkt->resent = 0;
	    pkt->seq_no = seq_no ++;
	    enq(node, pkt);
	}
#endif

	if (BIG_CTH_DEBUG)
	    printf("%f: test: tick %d, busy %d, out %d, touts %d timeout %f dup_acks %d\n", make_float(now), tick, line_busy, num_outstanding, num_timeouts, timeout, dup_ack_retx);

	if (tick 
	    and (not line_busy)
#ifndef NO_FLOW_CONTROL
	    and ((num_outstanding - 1 < ftp_window - tick)
		or num_timeouts or (dup_ack_retx >= 2))

	    and ((last_sent < last_ack + ftp_window - tick)
		or num_timeouts or (dup_ack_retx >= 2))

#endif
	    and (last_sent < num_pkts[node] -1)
	    )
	{
	    pkt = deq(node);
	    if(pkt is NULL)
		pr_error("in test, dequeued a null pkt!!!");

#ifndef NO_FLOW_CONTROL
	    tick--;
	    if (tick is 0)
	    {
		if (num_timeouts >= 2)
		    num_timeouts -= 2;
		else
		    num_timeouts = 0;
		if (dup_ack_retx >= 2)
		    dup_ack_retx -= 2;
		else
		    dup_ack_retx = 0;
	    }

	    /* set up packet pair ID: (tick=1) => decbit = 0 and vice versa */

	    pkt->decbit = 1 - tick;
#endif
	    pkt->gen_time = runtime();
	    pkt->id = transmission_id++;
	    if(pkt->resent)
		num_retransmissions[node]++;
	    else
            	num_retx[node][pkt->seq_no % MAX_WINDOW_SIZE] = 0;

	    recvd_ok[node][pkt->seq_no % MWS] = 0;
	    if(pkt->seq_no > last_sent)
		last_sent = pkt->seq_no;
	    safe_send(pkt, timeout);
	    if (PP_PRINT)
	        make_plot("num_out", num_outstanding);
	    num_outstanding++;
	    if (PP_PRINT)
	        make_plot("num_out", num_outstanding);
	    line_busy = 1;
	}
	goto recv;
    }
}
