/*
 * Nest simulation library - message routines
 *
 * $Header: messes.c,v 2.5 88/03/31 14:49:37 dupuy Rel $
 */

#include "nest.h"

#include "defs.h"
#include "simulate.h"
#include "message.h"
/* Global */

messlist **_message_table;       /* size of `nodes' arg to simulate() */

/* VARARGS 6 */
int 
reliable (a_node, sender, dest, key, dataptr, delay)
    ident           a_node,	/* actual destination node (never 0) */
                    sender,	/* sender (usually _current_node) */
                    dest;	/* destination node (0 if all nbrs) */
    long            key;	/* message code (user defined int) */
    char           *dataptr;	/* pointer to message data (on heap) */
    long            delay;	/* time at which message will arrive */
{
    messlist      **mlist, *newitem;
    message        *newmsg;


    if (dataptr is 0)
    {
	printf ("attempt to deliver null msg in reliable\n");
	printf ("sender %d, dest %d, a_node %d, _current_node %d\n",
		sender, dest, a_node, _current_node);
	exit (0);
    }
    if (a_node is 0 or a_node > Nodes)	/* a_node must be a valid node */
	return (0);

    report (("sent message (%d) from %d to %d[%d]\n",
	     key, sender, a_node, dest));

    /* create "message_unit" and "message_list" structures */

    newitem = new (messlist);
    newitem->msg = newmsg = new (message);

    /* set "message_unit" fields */

    newmsg->key = key;
    newmsg->dataptr = dataptr;
    newmsg->sender = sender;
    newmsg->dest = dest;

    /* update sent and arrived time and blocked status of new message */

    _reschedule (a_node, delay is 0 ? Delay : delay,
		 &(newmsg->sent), &(newmsg->arrived));

    /* insert message into message list for this node */

    mlist = &_message_table[a_node];
    while ((*mlist) isnt nil and
	   not time_after ((*mlist)->msg->arrived, newmsg->arrived))

	advancn (mlist);	/* search for later message */

    insn (newitem, mlist);	/* insert new message at this point */

    return (1);			/* currently, this always succeeds */
}



_drop_message (messlistp)

    messlist      **messlistp;	/* address of pointer to message */
{
    messlist       *messlist_rest;


    /* free the received message and remove from messlist */

    messlist_rest = (*messlistp)->next;

    dispose ((*messlistp)->msg);
    dispose (*messlistp);

    (*messlistp) = messlist_rest;
}


_message_init (nodes)
    unsigned        nodes;
{
    _message_table = newarray (nodes + 1, messlist *);
}


timev          *
_check_queue (node, fromtime, totime)
    ident           node;
    timev          *fromtime, *totime;
{
    messlist       *messages;

    messages = _message_table[node];

    foreach (messages)
    {
	if (time_after (*fromtime, messages->msg->arrived))
	    continue;

	if (totime isnt nil and time_before (*totime, messages->msg->arrived))
	    break;

	return (&messages->msg->arrived);
    }

    return (nil);
}
