/*
 * Nest simulation library - alternate receive message routine
 * 
 * $Header: recvmt.c,v 2.5 88/03/31 14:50:03 dupuy Rel $
 */

#include "nest.h"

#include "defs.h"
#include "simulate.h"
#include "message.h"
#include "process.h"

ident           recvmt (destp, keyp, dataptrp, sendt, arrvt)    /* returns sending node */
ident          *destp;                  /* gets dest of initial sendm() call */
long           *keyp;                   /* gets message code (user defined) */
char          **dataptrp;               /* gets ptr to message data on heap */
timev          *sendt;                  /* time at which message was sent */
timev          *arrvt;                  /* time at which message arrived */
{
    ident           sender;
    register message *msg_ptr;


    hold ();
    if (_stop_timer (time_zero))        /* stop the clock while we work */
        fatal ("recvm: stop_timer: ");

    report (("%d receiving message\n", _current_node));

    if (_message_table[_current_node] is nil)
    {                                   /* block indefinitely if no messages */
        if (_block (time_zero, RECEIVE))
            fatal ("recvm: block(0): ");
    }

    /* get the earliest message which exists (should be a separate function) */

    msg_ptr = _message_table[_current_node]->msg;

    if (time_after (msg_ptr->arrived, _proc_table[_current_node].runtime))
    {                                   /* block until that message arrives */
        if (_block (msg_ptr->arrived, RECEIVE))
            fatal ("recvm: block(1): ");
    }
    else
    {                                   /* block until present time */
        if (_block (_proc_table[_current_node].runtime, RECEIVE))
            fatal ("recvm: block(2): ");
    }


    /* get the first message which arrives (it may have changed) */

    msg_ptr = _message_table[_current_node]->msg;

    /* we reach here only if a message exists, and has arrived already */

    /* set up recvm's argument variables */

    sender = msg_ptr->sender;           /* temporary value for return() */

    if (destp)
        *destp = msg_ptr->dest;

    if (keyp)
        *keyp = msg_ptr->key;

    if (dataptrp)
        *dataptrp = msg_ptr->dataptr;

    if (sendt)
        *sendt = msg_ptr->sent;

    if (arrvt)
        *arrvt = msg_ptr->arrived;

    report (("%d got message [%d] from %d to %d\n",
             _current_node, *keyp, sender, *destp));

    /* remove the message from the message list */

    _drop_message (&_message_table[_current_node]);

    if (_start_timer (time_zero))       /* start the clock again */
        fatal ("recvm: start_timer: ");
    release (1);

    return (sender);                    /* return the message's sender */
}
