/*
 * Nest simulation library - send message routine
 * 
 * $Header: sendm.c,v 2.5 88/03/31 14:50:21 dupuy Rel $
 */

#include "nest.h"

#include "defs.h"
#include "simulate.h"
#include "message.h"
#include "network.h"
#include "process.h"


sendm (dest, key, dataptr)              /* sends a message; returns status */
ident           dest;                   /* destination node (0 for all nbrs) */
long            key;                    /* message code (user defined int) */
char           *dataptr;                /* pointer to message data (on heap) */
{
    int             status;
    neighbor       *nlist;
    funcptr        *stack;
    funcptr        *copy_stack ();

    hold ();
    _timers = false;
    nlist =                             /* set up pointer into neighbor list */
        _network_table[_current_node];

    if (Broadcast or (dest is 0))       /* send message to all its neighbors */
    {
        if (Point2Point)                /* not allowed to broadcast to 0 */

            status = Error;             /* return Error for bad destination */

        else
        {
            status = 0;

            foreach (nlist)             /* for each and every neighbor */
            {
                stack = copy_stack (nlist->chanstack);

                status +=               /* send each of them the message */
                    channel_sendm (nlist->nbr, _current_node,
                                   dest, key, dataptr, nlist->weight, stack);

                dispose (stack);
            }
        }
    }

    else                                /* send message only to destination */
    {
        while (nlist isnt nil and nlist->nbr isnt dest)
            nlist = nlist->next;        /* search for the destination node */

        if (nlist is nil)               /* destination node does not exist */

            status = Error;             /* return Error for bad destination */

        else                            /* send that node the message */
        {
            stack = copy_stack (nlist->chanstack);

            status = channel_sendm (nlist->nbr, _current_node,
                                    dest, key, dataptr, nlist->weight, stack);

            dispose (stack);
        }
    }

    if (_start_timer (time_zero))       /* start the clock again and return */
        fatal ("sendm: start_timer: ");
    release (1);

    return (status);                    /* return Error, or sum of channels */
}


static funcptr *copy_stack (stack)
chanptr         stack;
{
    funcptr        *channels;
    reg chanfunc   *func;
    int             count;

    count = 1;
    func = stack;
    foreach (func)
        count++;

    channels = newarray ((unsigned) count, funcptr);

    count = 0;
    func = stack;
    foreach (func)
        channels[count++] = func->addr;

    channels[count] = nil;

    return (channels);
}

/* should really be in mess2s.c */

_reschedule (nodeid, later, nowtime_ptr, arrivedtime_ptr)
ident           nodeid;
long            later;
timev          *nowtime_ptr,
               *arrivedtime_ptr;

{
    *nowtime_ptr = _proc_table[_current_node].runtime;
    *arrivedtime_ptr = *nowtime_ptr;

    /* arrived time is the current runtime plus the delay */
    arrivedtime_ptr->tv_sec +=
        (arrivedtime_ptr->tv_usec += later) / MILLION;
    arrivedtime_ptr->tv_usec %= MILLION;

    if (_proc_table[nodeid].whyblocked is RECEIVE
            or _proc_table[nodeid].whyblocked is WAIT
            or _proc_table[nodeid].whyblocked is RECV (_current_node))
    {
        if (time_iszero (_proc_table[nodeid].unblocked))
            /* remove from blocked queue */
            _remove_sched (nodeid, BlockedQueue);
        else
        {                               /* arrive before unblocked time */
            if (time_before (*arrivedtime_ptr,
                             _proc_table[nodeid].unblocked))
            {

                /*
                 * if unblocked less than sync_time, remove from waitqueue,
                 * otherwise, remove from newpass waitqueue
                 */

                if (time_before (_proc_table[nodeid].unblocked, _sync_time))
                    _remove_sched (nodeid, WaitQueue);
                else
                    _remove_sched (nodeid, NewpassWaitqueue);
            }
            else                        /* no rescheduling necessary */
                return;
        }
        _proc_table[nodeid].unblocked =
            when (time_after (*arrivedtime_ptr, _proc_table[nodeid].runtime))
            (
             then (*arrivedtime_ptr)
             otherwise (_proc_table[nodeid].runtime)
            );

        _insert_sched (nodeid);
    }
}
