/*
 * Nest simulation library - main simulation loop routines
 * 
 * $Header: simulate.c,v 2.5 88/03/31 14:50:25 dupuy Rel $
 */

#define D(x)	/* x */
#include "nest.h"
#include "graph.h"
/* #include "dccsig.h"*/

/*
 * Top level function of simulation layer; performs scheduling.
 */

#define XSize 8192                      /* size of temporary stack */
#include <errno.h>

#include "defs.h"
#include "process.h"
#include "schedule.h"
#include "simulate.h"

/*
 * Define nest internal global variables which are used throughout the
 * simulation layer.
 */


flag     _running;               /* running some node's function */
flag     _swapping;              /* `on swap stack' indication */
flag     _timers;                /* timer interrupts are enabled */
flag     _repeating;             /* timing forces repeat of node */

jmp_buf  _scheduler;             /* jump buffer for top of "loop" */
char    *_stackbase;             /* base of the swapped stack */
timev    _passtime;              /* length of simulation passes */
timev    _sync_time;             /* synchronization time of this pass */
int      _passes;
int      _runs;

int      errno;


int             simulate (nodes, stacksize, initgraph, port)
unsigned        nodes,                  /* number of nodes in graph */
                stacksize;              /* total size of all nodes' stacks */
graph          *initgraph;              /* initial network graph */
int             port;                   /* port number for simulation */
{
    Nodes = nodes;                      /* [re]initialize globals */
    StackSize = stacksize;
    PortNumber = port;

    /* select profile or virtual timer; set scheduler signal mask, handlers */

    /*_signal_init();			/* initialize trap handler for bus/io  errors */

    _sched_init (nodes);                /* all on done_list, n_id = node */

    nest_proc_init (nodes);                 /* _sync_time = 0.01 sec */

    _network_init (nodes);              /* no connections anywhere */

    _message_init (nodes);              /* no messages for anybody   */

    /* create server port, no connections; set initial network configuration */

    if (_serv_init (nodes, initgraph, port))
        return (_complain (""));

    if (_swap_init (nodes, stacksize))  /* set up area for storing stacks */
        return (_complain (""));

    _passes = _runs = 0;                /* start counters at 0 */

    _timers = _running = false;

    return (s_main ());                 /* end of initialization */
}

static int      s_main ()               /* main loop of simulation */
{
    char            tmpstack[XSize];
    int             state ;
    reg ident       next_node;


    /* top of "loop" created by setjmp() and longjmp() */


    _stackbase = nest_stack ();
    state = setjmp (_scheduler);        /* save simulator state for return */

    _repeating = false;                 /* cleared at top of each "loop" */
    if (state is NewPass or state is 0) /* we are starting a new pass or first time*/
    {
        _serv_exec ();		/* call the monitor for the first time */
        time_add (_sync_time, _passtime);
        _passes++;

        if (_next_queue is nil)
            if (_next_waitqueue is nil)
            {                           /* we have deadlock or are done */
                next_node = 0;
            }
            else                        /* we have only waiting nodes */
                next_node = _next_pass (_next_waitqueue);

        else                            /* move to new execution queue */
            next_node = _next_pass (_next_queue);
    }
    else                                /* we are continuing a pass */
    {
        next_node = _get_next_node (_current_node);
        if (_proc_table[_current_node].whyblocked is UNBLOCKED)
        {
            /* stopped by a timeout; update _sched_table */

            _newpass_add (_current_node);
        }
        else                            /* stopped because of a block */
        {
            _insert_sched (_current_node);

            if ((next_node is 0 and _wait_queue isnt nil) or
                    (next_node isnt 0 and
                     _sched_table[next_node].prev isnt nil))
            {
                report (("repeating node %d\n", _current_node));
                _repeating = true;      /* repeating node on wait queue */
                next_node = _wait_queue->n_id;
            }                           /* this catches nodes blocking us */
        }

        if (next_node is 0)             /* at end of queue, start a new pass */
            longjmp (_scheduler, NewPass);
    }


    /* top of short loop where nodes never timeout or block, but simply end */

    while (state isnt Error and next_node isnt 0)
    {
        reg process    *proc = &(_proc_table[next_node]);

        new_current_node (next_node);

        if (_proc_table[_current_node].started is false)     
				/* then start function on node */
        {
            if (_start_timer ((_proc_table[_current_node]).unblocked))
                fatal ("simulate: start_timer: ");

            nest_proc_run (_current_node, &(_proc_table[_current_node]));

            if (_timers )
		_timers = false; 

            next_node = _get_next_node (_current_node);

            if (_proc_table[_current_node].repeat is true)
                nest_proc_start (_current_node);
            else
                _done_queue = _push_node (_current_node, _done_queue);

            if (next_node is 0)         /* at end of queue, start a new pass */
                longjmp (_scheduler, NewPass);
        }
        else
        {

            /*
             * if `proc' isn't a true register, you may need to replace it
             * with `_proc_table[_current_node]' in this block (see stack.s).
             */

            _swapping = true;           /* switch stack to avoid conflicts */

#if STACKDIR < 1
            nest_xstack (tmpstack + sizeof (tmpstack) - 256);
#else                                   /* -256 above in case we postdecrement */
            nest_xstack (tmpstack);
#endif
            /* copy its stack back and dealloc */

            if (nest_restore (_current_node))
                fatal ("simulate: restore: ");

            nest_cook (&(_proc_table[_current_node].environment)); /* start running it again */

            fatal ("simulate: cook");
        }
    }


    /*
     * Simulation has finished
     */

    return (state);                     /* Error or None */
}




static          new_current_node (nodeid)
ident           nodeid;
{
    ++_runs;
    _current_node = nodeid;

    report ((_show_sched ("execution_queue", &_sched_table[nodeid])));

    if (Logging)
        (void) printf ("running node %d  \r", nodeid);
}
