/*
 * Nest simulation library - timer routines
 * 
 * $Header: timers.c,v 2.5 88/03/31 14:50:46 dupuy Rel $
 */

#include "nest.h"
/* #include "dccsig.h"*/
#include "defs.h"
#include "simulate.h"
#include "process.h"


/*
 * Blocks a node by generating a timer signal. Sets the unblocked time of this
 * node to the `unblocktime', and the reason for blocking to `reason'. Will
 * return only after process runtime is at least unblocktime (if unblocktime
 * value is zero, it may never return).
 */

_block (unblocktime, reason)
timev           unblocktime;
int             reason;
{
    process *proc = &(_proc_table[_current_node]);
    extern int      errno;
    int             olderrno;
    pointer         stack;


    if ((proc->whyblocked = reason)     /* no blocking without a reason: */
            is UNBLOCKED)               /* (scheduler would start timer) */
        return (trap (_error ("node blocked without reason")));

    proc->unblocked = unblocktime;

    _timers = _running = false;         /* necessary for proper malloc/free */

    olderrno = errno;                   /* save this global, anyhow */

    stack = nest_stack ();
    if (nest_store (stack, 0))
        fatal ("_block: store: ");

    if(not nest_freeze (&proc->environment))
        longjmp (_scheduler, Normal);       /* stop running this node */

    _swapping = false;

    if (proc->whyblocked is UNBLOCKED)
        if (_start_timer (proc->unblocked))
            fatal ("timeout: start_timer: ");
    errno = olderrno;                   /* restore this global variable */

    _running = true;                    /* resume running this node */


    if (proc->whyblocked is ADVANCE)    /* eat up fake cpu time */
        time_elapsed_add (proc->cputime, proc->unblocked, proc->runtime);

    proc->runtime = proc->unblocked;    /* set runtime to unblocked time */
    proc->timeleft = _sync_time;        /* set timeleft to synctime-runtime */
    time_sub (proc->timeleft, proc->runtime);
    proc->whyblocked = UNBLOCKED;

    return (None);
}


/*
 * Starts profile and alarm timers. If nowtime isn't time_zero, sets runtime
 * for this node to nowtime and calculates time left for this node. Sets alarm
 * timer to wakeups interval, and profile timer to the time left for this
 * node.
 */

_start_timer (nowtime)
timev           nowtime;
{
    process *proc = &(_proc_table[_current_node]);

    if (time_nonzero (nowtime))         /* we should set new time values */
    {
        proc->runtime = nowtime;
        proc->timeleft = _sync_time;
        time_sub (proc->timeleft, nowtime);

    }
    else
    {
        if (--proc->stopcount isnt 0)   /* check for stop_time nesting */
            return (None);

        while (not time_after (proc->timeleft, time_zero))
        {                               /* if time left is already used up */
            if (_block (proc->runtime, SLEEP))
                return (_error ("block: "));
        }                               /* block til all time used is past */
    }

    _timers = true;
    return (None);
}




_stop_timer (addtime)
timev           addtime;
{
    reg process    *proc = &(_proc_table[_current_node]);
    timev           elapsedtime;

    _timers = false;

    if (time_nonzero (addtime))         /* non positive addtime is an error */
        if (time_negative (addtime))
            return (trap (_error ("negative addtime specified")));

    /* increment runtime by addtime argument */

    time_add (proc->runtime, addtime);
    time_sub (proc->timeleft, addtime);

    return (None);                      /* return 0 if no error occurs */
}

